"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.construct = exports.constructClassName = exports.constructMods = exports.constructElem = exports.constructBlock = void 0;
const types = __importStar(require("@babel/types"));
const constants_1 = require("./constants");
const types_1 = require("./types");
function* constructBlock({ block }, shouldSkip) {
    if (shouldSkip) {
        return;
    }
    if (!block.length || shouldSkip) {
        yield constants_1.EMPTY;
    }
    if ((0, types_1.isArray)(block)) {
        for (const { value } of block) {
            if (!value || typeof value !== 'string') {
                continue;
            }
            yield value;
        }
    }
    if (typeof block === 'string') {
        yield block;
    }
}
exports.constructBlock = constructBlock;
;
function* constructElem(bemProps) {
    const { block, elem } = bemProps;
    if (block.length === 0 || !elem || elem.length === 0) {
        return constants_1.EMPTY;
    }
    for (const _block of constructBlock(bemProps)) {
        if (!_block) {
            continue;
        }
        if ((0, types_1.isArray)(elem)) {
            for (const { value: _elem } of elem) {
                if (!_elem) {
                    continue;
                }
                yield `${_block}${(0, constants_1.ELEM_CONNECTOR)()}${_elem}`;
            }
        }
        if (typeof elem === 'string' && elem) {
            yield `${_block}${(0, constants_1.ELEM_CONNECTOR)()}${elem}`;
        }
    }
}
exports.constructElem = constructElem;
;
function* constructMods(bemProps, isBlockTopLevel) {
    const { elem, mods } = bemProps;
    if (!isBlockTopLevel && !(elem === null || elem === void 0 ? void 0 : elem.length)) {
        return constants_1.EMPTY;
    }
    // If mods is empty string or
    if (!mods || (!(elem === null || elem === void 0 ? void 0 : elem.length) && !((mods === null || mods === void 0 ? void 0 : mods.length) || Object.keys(mods).length))) {
        return constants_1.EMPTY;
    }
    // If we have 'elem', use that, otherwise use 'block'
    const modsPrefixIterator = (elem === null || elem === void 0 ? void 0 : elem.length)
        ? constructElem(bemProps)
        : constructBlock(bemProps);
    const prefixes = [];
    for (const prefix of modsPrefixIterator) {
        prefixes.push(prefix);
    }
    if ((0, types_1.isArray)(mods)) {
        for (const mod of mods) {
            if (types.isStringLiteral(mod) && mod.value) {
                yield prefixes.reduce((acc, prefix) => {
                    const SPACE_AFTER_ACC = acc && prefix
                        ? constants_1.WHITESPACE
                        : constants_1.EMPTY;
                    return prefix
                        ? `${acc}${SPACE_AFTER_ACC}${prefix}${(0, constants_1.MODS_CONNECTOR)()}${mod.value}`
                        : constants_1.EMPTY;
                }, constants_1.EMPTY);
            }
            if (types.isObjectProperty(mod)) {
                const { key, value } = mod;
                const modName = types.isIdentifier(key)
                    ? key.name
                    : types.isStringLiteral(key)
                        ? key.value
                        : constants_1.EMPTY;
                // When the right hand side of the object property is a boolean and is false,
                // omit it altogether
                if (types.isBooleanLiteral(value) && !value.value) {
                    continue;
                }
                yield prefixes.reduce((acc, prefix) => {
                    const SPACE_AFTER_ACC = acc && prefix
                        ? constants_1.WHITESPACE
                        : constants_1.EMPTY;
                    return prefix
                        ? `${acc}${SPACE_AFTER_ACC}${prefix}${(0, constants_1.MODS_CONNECTOR)()}${modName}`
                        : constants_1.EMPTY;
                }, constants_1.EMPTY);
            }
        }
    }
    if (prefixes.length && typeof mods === 'string') {
        yield prefixes.reduce((acc, prefix) => {
            const SPACE_AFTER_ACC = acc && prefix
                ? constants_1.WHITESPACE
                : constants_1.EMPTY;
            return prefix
                ? `${acc}${SPACE_AFTER_ACC}${prefix}${(0, constants_1.MODS_CONNECTOR)()}${mods}`
                : constants_1.EMPTY;
        }, constants_1.EMPTY);
    }
    return constants_1.EMPTY;
}
exports.constructMods = constructMods;
const constructClassName = (bemProps) => {
    const { className } = bemProps;
    if (!className) {
        return constants_1.EMPTY;
    }
    if ((0, types_1.isArray)(className)) {
        return className.reduce((constructedClassName, currentClassName) => {
            const SPACE_AFTER_CLASSNAME = constructedClassName
                ? constants_1.WHITESPACE
                : constants_1.EMPTY;
            return `${constructedClassName}${SPACE_AFTER_CLASSNAME}${currentClassName.value}`;
        }, constants_1.EMPTY);
    }
    if (typeof className === 'object') {
        return constants_1.EMPTY;
    }
    return className;
};
exports.constructClassName = constructClassName;
const construct = (bemProps, isBlockTopLevel) => {
    const _className = (0, exports.constructClassName)(bemProps);
    const _conditionalExpressions = [];
    let _block = constants_1.EMPTY;
    let _elem = constants_1.EMPTY;
    let _mods = constants_1.EMPTY;
    if (isBlockTopLevel) { // If block is inherited, we don't want it in the className by itself
        for (const block of constructBlock(bemProps, !isBlockTopLevel)) {
            const SPACE = _block ? constants_1.WHITESPACE : constants_1.EMPTY;
            _block = `${_block}${SPACE}${block}`;
        }
    }
    for (const elem of constructElem(bemProps)) {
        const SPACE = _elem ? constants_1.WHITESPACE : constants_1.EMPTY;
        _elem = `${_elem}${SPACE}${elem}`;
    }
    for (const mod of constructMods(bemProps, isBlockTopLevel)) {
        const SPACE = _mods ? constants_1.WHITESPACE : constants_1.EMPTY;
        _mods = `${_mods}${SPACE}${mod}`;
    }
    if ((0, types_1.isArray)(bemProps.mods)) {
        const modsIterator = constructMods(bemProps, isBlockTopLevel);
        for (const mod of bemProps.mods) {
            if (types.isObjectProperty(mod)) {
                const { value } = mod;
                // When the right hand side of the object property is a boolean, it can only be always true or always false
                // so there is no point in constructing a conditional expression
                if (types.isBooleanLiteral(value)) {
                    continue;
                }
                const conditionalExpression = getConditionalExpression(mod, modsIterator);
                if (!conditionalExpression) {
                    continue;
                }
                _conditionalExpressions.push(conditionalExpression);
            }
        }
    }
    const SPACE_AFTER_BLOCK = _block && (_elem || _mods || _className)
        ? constants_1.WHITESPACE
        : constants_1.EMPTY;
    const SPACE_AFTER_ELEM = _elem && (_className || _mods) && !(_conditionalExpressions.length > 0)
        ? constants_1.WHITESPACE
        : constants_1.EMPTY;
    const SPACE_AFTER_MODS = _mods && _className
        ? constants_1.WHITESPACE
        : constants_1.EMPTY;
    const SPACE_AFTER_CLASSNAME = _className
        ? constants_1.WHITESPACE
        : constants_1.EMPTY;
    const CLASS_NAME = _conditionalExpressions.length
        ? `${_block}${SPACE_AFTER_BLOCK}${_elem}${SPACE_AFTER_ELEM}${_className}${SPACE_AFTER_CLASSNAME}`
        : `${_block}${SPACE_AFTER_BLOCK}${_elem}${SPACE_AFTER_ELEM}${_mods}${SPACE_AFTER_MODS}${_className}`;
    // If we have conditional expresssions, construct template literals using them
    if (_conditionalExpressions.length) {
        return (types.jsxAttribute(types.jsxIdentifier(types_1.BEMPropTypes.CLASSNAME), types.jsxExpressionContainer(types.templateLiteral([
            types.templateElement({ raw: CLASS_NAME }, false),
            ..._conditionalExpressions.map(() => types.templateElement({ raw: '' }, false))
        ], _conditionalExpressions))));
    }
    // By default, we'll use a simple string literal
    return (types.jsxAttribute(types.jsxIdentifier(types_1.BEMPropTypes.CLASSNAME), types.stringLiteral(CLASS_NAME)));
};
exports.construct = construct;
const getConditionalExpression = (mod, modsIterator) => {
    if (!mod) {
        return;
    }
    const modValue = modsIterator.next().value;
    if (!modValue) {
        return;
    }
    return types.conditionalExpression(mod.value, types.stringLiteral(' ' + modValue), types.stringLiteral(constants_1.EMPTY));
};
exports.default = exports.construct;
