"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const types = __importStar(require("@babel/types"));
const types_1 = require("./types");
const constants_1 = require("./constants");
const construct_1 = __importDefault(require("./construct"));
function default_1() {
    return {
        name: 'transform-bem-props',
        visitor: {
            JSXElement(element, state) {
                var _a;
                if ((_a = state.file.ast.comments) === null || _a === void 0 ? void 0 : _a.some((comment) => comment.value.includes('bem-attr-skip-file'))) {
                    return;
                }
                traverseJSXElementTree(element, constants_1.EMPTY);
                // Don't traverse child nodes, as we will do that manually
                element.skipKey('children');
            }
        }
    };
}
exports.default = default_1;
;
/**
 * Recursively traverses the JSXElement tree and constructs the 'className' attribute
 * @param element - The JSXElement to recursively traverse
 * @param block - Passed to the next iteration to allow block inheritance
 */
const traverseJSXElementTree = (element, block) => {
    // const comments = element.parentPath.node.leadingComments?.join(element.parentPath.node.innerComments) ;
    const comments = [];
    element.parentPath.node.leadingComments && comments.push(...element.parentPath.node.leadingComments);
    element.parentPath.node.innerComments && comments.push(...element.parentPath.node.innerComments);
    element.parentPath.node.trailingComments && comments.push(...element.parentPath.node.trailingComments);
    element.node.innerComments && comments.push(...element.node.innerComments);
    element.node.leadingComments && comments.push(...element.node.leadingComments);
    element.node.trailingComments && comments.push(...element.node.trailingComments);
    if (comments.length && comments.some((comment) => comment.value.includes('no-traverse'))) {
        return;
    }
    ;
    const { node: { openingElement: { attributes, loc, name: htmlTagName } } } = element;
    const attributeIndexesToRemove = [];
    const bemProps = {
        block,
        elem: constants_1.EMPTY,
        mods: constants_1.EMPTY,
        className: constants_1.EMPTY
    };
    let hasFoundValidBlock = false;
    attributes.forEach((attribute, index) => {
        if (types.isJSXSpreadAttribute(attribute)) {
            return;
        }
        const { value: valueNode, name: { name } } = attribute;
        if (!constants_1.BEM_PROP_TYPES.includes(name)) {
            return;
        }
        if (types.isStringLiteral(valueNode)) {
            const { value } = valueNode;
            if (value) {
                bemProps[name] = value;
                if (name === types_1.BEMPropTypes.BLOCK && value) {
                    hasFoundValidBlock = true;
                }
            }
        }
        // attribute={/* this is a JSX expression container */}
        if (types.isJSXExpressionContainer(valueNode)) {
            const { expression } = valueNode;
            if (types.isStringLiteral(expression)) {
                const { value } = expression;
                if (value) {
                    bemProps[name] = value;
                    if (name === types_1.BEMPropTypes.BLOCK && value) {
                        hasFoundValidBlock = true;
                    }
                }
            }
            if (types.isArrayExpression(expression)) {
                const { elements } = expression;
                if (elements.length) {
                    bemProps[name] = elements;
                    if (name === types_1.BEMPropTypes.BLOCK && elements.length) {
                        hasFoundValidBlock = true;
                    }
                }
            }
            if (types.isObjectExpression(expression) && name === types_1.BEMPropTypes.MODS) {
                const { properties } = expression;
                bemProps.mods = properties;
            }
        }
        attributeIndexesToRemove.push(index);
    });
    // If there was no new 'block' defined on the element, but 'elem' or 'mods' were
    if (!hasFoundValidBlock && (bemProps.elem || bemProps.mods)
        && (0, constants_1.DISABLE_BLOCK_INHERITANCE)()) {
        handleUndefinedBlock(bemProps.block, htmlTagName, loc);
    }
    // Remove all attributes that were processed.
    // The reason for not removing it directly in the loop
    // is that it messes up the indexes of the attributes, leading to skipped elements.
    const attributePaths = element.get('openingElement.attributes');
    attributeIndexesToRemove.forEach(attributeIndex => {
        attributePaths[attributeIndex].remove();
    });
    // if ((bemProps.block && hasFoundValidBlock) || bemProps.elem || bemProps.mods) {
    // }
    // When block inheritance is disabled, we will need to have defined on every line a new block,
    // therefore hasFoundValidBlock would be always true. So, to correctly check if block is top level,
    // compare the newly assigned block to the one from the previous iteration.
    // (if they match => they are the same, which means we have in fact *not* found a new block)
    // If you are trying to understand what is happening below, please forgive me :|
    const isBlockTopLevelEvaluation = typeof bemProps.block === 'string'
        ? (hasFoundValidBlock && bemProps.block !== block)
        : (0, types_1.isArray)(bemProps.block) && (0, types_1.isArray)(block) // When both are arrays, compare them
            ? hasFoundValidBlock && JSON.stringify(bemProps.block) === JSON.stringify(block)
            : ( // When they have differing types, we know block is top level
            (0, types_1.isArray)(bemProps.block) && typeof block === 'string'
                || (0, types_1.isArray)(block) && typeof bemProps.block === 'string');
    const isBlockTopLevel = (0, constants_1.DISABLE_BLOCK_INHERITANCE)()
        ? isBlockTopLevelEvaluation || (!block && hasFoundValidBlock)
        : hasFoundValidBlock;
    const classNameAttribute = (0, construct_1.default)(bemProps, isBlockTopLevel);
    if (classNameAttribute) {
        const { value } = classNameAttribute;
        if ((types.isStringLiteral(value) && value.value)
            || types.isJSXExpressionContainer(value)) {
            attributes.push(classNameAttribute);
        }
    }
    element.get('children').forEach(childElement => {
        if (!types.isJSXElement(childElement)) {
            return;
        }
        traverseJSXElementTree(childElement, bemProps.block);
    });
};
const handleUndefinedBlock = (block, htmlTagName, location) => {
    const { name } = htmlTagName;
    const { start: { line, column } = {
        line: 'unknown',
        column: 'unknown'
    }, } = location || {};
    const reducer = (acc, value) => {
        const SEPARATOR = (acc && value.value)
            ? `${constants_1.COMMA}${constants_1.WHITESPACE}`
            : constants_1.EMPTY;
        return `${acc}${SEPARATOR}${value.value}`;
    };
    const inheritedBlock = (0, types_1.isArray)(block)
        ? block.reduce(reducer, constants_1.EMPTY)
        : block;
    const inheritedMessage = inheritedBlock
        ? `Inherited [${inheritedBlock}], but block inheritance is disabled.`
        : 'Did not inherit from parent.';
    const message = `Block is not defined on <${name}> at line ${line}, column ${column}. ${inheritedMessage}`;
    throw Error(message);
};
